@@ -14,6 +14,7 @@ import itertools
 import logging
 import os
 import platform
+import shutil
 import stat
 import sys
 import typing
@@ -28,13 +29,21 @@ except ImportError:
     except ImportError:  # pragma: no cover
         scandir = None
 
+try:
+    from os import sendfile
+except ImportError:
+    try:
+        from sendfile import sendfile  # type: ignore
+    except ImportError:
+        sendfile = None
+
 from . import errors
 from .errors import FileExists
 from .base import FS
 from .enums import ResourceType
 from ._fscompat import fsencode, fsdecode, fspath
 from .info import Info
-from .path import basename
+from .path import basename, dirname
 from .permissions import Permissions
 from .error_tools import convert_os_errors
 from .mode import Mode, validate_open_mode
@@ -107,7 +116,7 @@ class OSFS(FS):
             root_path = fsdecode(root_path)
         self.root_path = root_path
         _drive, _root_path = os.path.splitdrive(fsdecode(fspath(root_path)))
-        _root_path = _drive + (_root_path or '/') if _drive else _root_path
+        _root_path = _drive + (_root_path or "/") if _drive else _root_path
         _root_path = os.path.expanduser(os.path.expandvars(_root_path))
         _root_path = os.path.normpath(os.path.abspath(_root_path))
         self._root_path = _root_path
@@ -258,7 +267,7 @@ class OSFS(FS):
         with convert_os_errors("getinfo", path):
             _stat = os.stat(fsencode(sys_path))
             if "lstat" in namespaces:
-                _stat = os.lstat(fsencode(sys_path))
+                _lstat = os.lstat(fsencode(sys_path))
 
         info = {
             "basic": {"name": basename(_path), "is_dir": stat.S_ISDIR(_stat.st_mode)}
@@ -321,8 +330,8 @@ class OSFS(FS):
         _path = self.validatepath(path)
         sys_path = self._to_sys_path(_path)
         with convert_os_errors("openbin", path):
-            if six.PY2 and _mode.exclusive and self.exists(path):
-                raise errors.FileExists(path)
+            if six.PY2 and _mode.exclusive:
+                sys_path = os.open(sys_path, os.O_RDWR | os.O_CREAT | os.O_EXCL)
             binary_file = io.open(
                 sys_path, mode=_mode.to_platform_bin(), buffering=buffering, **options
             )
@@ -361,90 +370,82 @@ class OSFS(FS):
     # Optional Methods
     # --------------------------------------------------------
 
+    # --- Type hint for opendir ------------------------------
+
     if False:  # typing.TYPE_CHECKING
 
         def opendir(self, path, factory=None):
             # type: (_O, Text, Optional[_OpendirFactory]) -> SubFS[_O]
             pass
 
-    def getsyspath(self, path):
-        # type: (Text) -> Text
-        sys_path = os.path.join(self._root_path, path.lstrip("/").replace("/", os.sep))
-        return sys_path
+    # --- Backport of os.sendfile for Python < 3.8 -----------
 
-    def geturl(self, path, purpose="download"):
-        # type: (Text, Text) -> Text
-        if purpose != "download":
-            raise NoURL(path, purpose)
-        return "file://" + self.getsyspath(path)
+    def _check_copy(self, src_path, dst_path, overwrite=False):
+        # validate individual paths
+        _src_path = self.validatepath(src_path)
+        _dst_path = self.validatepath(dst_path)
+        # check src_path exists and is a file
+        if self.gettype(src_path) is not ResourceType.file:
+            raise errors.FileExpected(src_path)
+        # check dst_path does not exist if we are not overwriting
+        if not overwrite and self.exists(_dst_path):
+            raise errors.DestinationExists(dst_path)
+        # check parent dir of _dst_path exists and is a directory
+        if self.gettype(dirname(dst_path)) is not ResourceType.directory:
+            raise errors.DirectoryExpected(dirname(dst_path))
+        return _src_path, _dst_path
 
-    def gettype(self, path):
-        # type: (Text) -> ResourceType
-        self.check()
-        sys_path = self._to_sys_path(path)
-        with convert_os_errors("gettype", path):
-            stat = os.stat(sys_path)
-        resource_type = self._get_type_from_stat(stat)
-        return resource_type
+    if sys.version_info[:2] < (3, 8) and sendfile is not None:
 
-    def islink(self, path):
-        # type: (Text) -> bool
-        self.check()
-        _path = self.validatepath(path)
-        sys_path = self._to_sys_path(_path)
-        if not self.exists(path):
-            raise errors.ResourceNotFound(path)
-        with convert_os_errors("islink", path):
-            return os.path.islink(sys_path)
+        _sendfile_error_codes = frozenset(
+            {
+                errno.EIO,
+                errno.EINVAL,
+                errno.ENOSYS,
+                errno.ENOTSUP,  # type: ignore
+                errno.EBADF,
+                errno.ENOTSOCK,
+                errno.EOPNOTSUPP,
+            }
+        )
 
-    def open(
-        self,
-        path,  # type: Text
-        mode="r",  # type: Text
-        buffering=-1,  # type: int
-        encoding=None,  # type: Optional[Text]
-        errors=None,  # type: Optional[Text]
-        newline="",  # type: Text
-        line_buffering=False,  # type: bool
-        **options  # type: Any
-    ):
-        # type: (...) -> IO
-        _mode = Mode(mode)
-        validate_open_mode(mode)
-        self.check()
-        _path = self.validatepath(path)
-        sys_path = self._to_sys_path(_path)
-        with convert_os_errors("open", path):
-            if six.PY2 and _mode.exclusive:
-                sys_path = os.open(sys_path, os.O_CREAT | os.O_EXCL)
-            _encoding = encoding or "utf-8"
-            return io.open(
-                sys_path,
-                mode=_mode.to_platform(),
-                buffering=buffering,
-                encoding=None if _mode.binary else _encoding,
-                errors=errors,
-                newline=None if _mode.binary else newline,
-                **options
-            )
+        def copy(self, src_path, dst_path, overwrite=False):
+            # type: (Text, Text, bool) -> None
+            with self._lock:
+                # validate and canonicalise paths
+                _src_path, _dst_path = self._check_copy(src_path, dst_path, overwrite)
+                _src_sys, _dst_sys = (
+                    self.getsyspath(_src_path),
+                    self.getsyspath(_dst_path),
+                )
+                # attempt using sendfile
+                try:
+                    # initialise variables to pass to sendfile
+                    # open files to obtain a file descriptor
+                    with io.open(_src_sys, "r") as src:
+                        with io.open(_dst_sys, "w") as dst:
+                            fd_src, fd_dst = src.fileno(), dst.fileno()
+                            sent = maxsize = os.fstat(fd_src).st_size
+                            offset = 0
+                            while sent > 0:
+                                sent = sendfile(fd_dst, fd_src, offset, maxsize)
+                                offset += sent
+                except OSError as e:
+                    # the error is not a simple "sendfile not supported" error
+                    if e.errno not in self._sendfile_error_codes:
+                        raise
+                    # fallback using the shutil implementation
+                    shutil.copy2(_src_sys, _dst_sys)
 
-    def setinfo(self, path, info):
-        # type: (Text, RawInfo) -> None
-        self.check()
-        _path = self.validatepath(path)
-        sys_path = self._to_sys_path(_path)
-        if not os.path.exists(sys_path):
-            raise errors.ResourceNotFound(path)
-        if "details" in info:
-            details = info["details"]
-            if "accessed" in details or "modified" in details:
-                _accessed = typing.cast(int, details.get("accessed"))
-                _modified = typing.cast(int, details.get("modified", _accessed))
-                accessed = int(_modified if _accessed is None else _accessed)
-                modified = int(_modified)
-                if accessed is not None or modified is not None:
-                    with convert_os_errors("setinfo", path):
-                        os.utime(sys_path, (accessed, modified))
+    else:
+
+        def copy(self, src_path, dst_path, overwrite=False):
+            # type: (Text, Text, bool) -> None
+            with self._lock:
+                _src_path, _dst_path = self._check_copy(src_path, dst_path, overwrite)
+                shutil.copy2(self.getsyspath(_src_path), self.getsyspath(_dst_path))
+
+    # --- Backport of os.scandir for Python < 3.5 ------------
 
     if scandir:
 
@@ -545,3 +546,84 @@ class OSFS(FS):
             start, end = page
             iter_info = itertools.islice(iter_info, start, end)
         return iter_info
+
+    # --- Miscellaneous --------------------------------------
+
+    def getsyspath(self, path):
+        # type: (Text) -> Text
+        sys_path = os.path.join(self._root_path, path.lstrip("/").replace("/", os.sep))
+        return sys_path
+
+    def geturl(self, path, purpose="download"):
+        # type: (Text, Text) -> Text
+        if purpose != "download":
+            raise NoURL(path, purpose)
+        return "file://" + self.getsyspath(path)
+
+    def gettype(self, path):
+        # type: (Text) -> ResourceType
+        self.check()
+        sys_path = self._to_sys_path(path)
+        with convert_os_errors("gettype", path):
+            stat = os.stat(sys_path)
+        resource_type = self._get_type_from_stat(stat)
+        return resource_type
+
+    def islink(self, path):
+        # type: (Text) -> bool
+        self.check()
+        _path = self.validatepath(path)
+        sys_path = self._to_sys_path(_path)
+        if not self.exists(path):
+            raise errors.ResourceNotFound(path)
+        with convert_os_errors("islink", path):
+            return os.path.islink(sys_path)
+
+    def open(
+        self,
+        path,  # type: Text
+        mode="r",  # type: Text
+        buffering=-1,  # type: int
+        encoding=None,  # type: Optional[Text]
+        errors=None,  # type: Optional[Text]
+        newline="",  # type: Text
+        line_buffering=False,  # type: bool
+        **options  # type: Any
+    ):
+        # type: (...) -> IO
+        _mode = Mode(mode)
+        validate_open_mode(mode)
+        self.check()
+        _path = self.validatepath(path)
+        sys_path = self._to_sys_path(_path)
+        with convert_os_errors("open", path):
+            if six.PY2 and _mode.exclusive:
+                sys_path = os.open(sys_path, os.O_RDWR | os.O_CREAT | os.O_EXCL)
+            _encoding = encoding or "utf-8"
+            return io.open(
+                sys_path,
+                mode=_mode.to_platform(),
+                buffering=buffering,
+                encoding=None if _mode.binary else _encoding,
+                errors=errors,
+                newline=None if _mode.binary else newline,
+                **options
+            )
+
+    def setinfo(self, path, info):
+        # type: (Text, RawInfo) -> None
+        self.check()
+        _path = self.validatepath(path)
+        sys_path = self._to_sys_path(_path)
+        if not os.path.exists(sys_path):
+            raise errors.ResourceNotFound(path)
+        if "details" in info:
+            details = info["details"]
+            if "accessed" in details or "modified" in details:
+                _accessed = typing.cast(int, details.get("accessed"))
+                _modified = typing.cast(int, details.get("modified", _accessed))
+                accessed = int(_modified if _accessed is None else _accessed)
+                modified = int(_modified)
+                if accessed is not None or modified is not None:
+                    with convert_os_errors("setinfo", path):
+                        os.utime(sys_path, (accessed, modified))
